package core

import (
	"fmt"
	"math/big"
	"math/rand"
	"sort"
	"sync"
	"time"
	"witness/conf"
	"witness/tree"
	"witness/util"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/ethdb/leveldb"
	"github.com/ethereum/go-ethereum/trie"
	witnessv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/witness/v1"
	log "github.com/sirupsen/logrus"
)

type Witness struct {
	db                 *trie.Database
	state              *StateDB
	pendingProof       map[common.Address]*witnessv1.ValidatedProof
	pendingWorkload    uint64
	mtTreeCache        map[string]*tree.MerkleTree
	date               string
	rpc                *ChainRPC
	containerAddresses []common.Address
	nmAddresses        []common.Address

	sync.Mutex
}

func NewWitness(cfg *conf.Config) *Witness {
	log.Infoln("######", cfg.PrivateKey)
	_rpc := newChain(cfg.ChainRPC, cfg.PrivateKey, cfg.StoreContract, cfg.WitnessContract)
	diskDB, err := leveldb.New(cfg.DataDir, 128, 1024, "", false)
	if err != nil {
		panic(err)
	}
	var lastDay string
	ok, err := diskDB.Has([]byte("lastday"))
	if err != nil {
		panic(err)
	}
	var stateRoot common.Hash
	if ok {
		val, _ := diskDB.Get([]byte("lastday"))
		log.WithField("last day", string(val)).Info("diskdb get last day")
		lastDay = string(val)
		root, _ := diskDB.Get([]byte(fmt.Sprintf("sroot:%s", string(val))))
		stateRoot = common.BytesToHash(root)
	}

	trieDB := trie.NewDatabase(diskDB)
	log.WithField("root", stateRoot).Info("new witness open trie")
	stateDB, err := NewStateDB(trieDB, stateRoot)
	if err != nil {
		panic(err)
	}
	w := &Witness{
		db:           trieDB,
		state:        stateDB,
		pendingProof: make(map[common.Address]*witnessv1.ValidatedProof),
		mtTreeCache:  make(map[string]*tree.MerkleTree),
		rpc:          _rpc,
		date:         lastDay,
	}
	w.LoadMerkleProof(lastDay)
	go w.UpdateAddress()
	<-time.After(time.Second * 3)
	go w.Ticker()
	// go w.Mock()

	return w
}

func (w *Witness) AddPendingProof(miner common.Address, proofs []*witnessv1.ValidatedProof) {
	w.Lock()
	defer w.Unlock()
	var workload uint64
	var pendingWorkload uint64

	currentPendingProof, ok := w.pendingProof[miner]
	if ok {
		workload = currentPendingProof.Workload
		pendingWorkload = currentPendingProof.Workload
	}
	for _, proof := range proofs {
		workload += proof.Workload
		w.pendingWorkload += proof.Workload
	}
	w.pendingProof[miner] = &witnessv1.ValidatedProof{Workload: workload}
	log.WithFields(log.Fields{
		"miner":            miner.Hex(),
		"proof_count":      len(proofs),
		"current_workload": workload,
		"pending_workload": pendingWorkload,
		"workload_diff":    workload - pendingWorkload,
	}).Debug("add validated proof")
}

func (w *Witness) ProcessDay() {
	w.date = time.Now().UTC().AddDate(0, 0, -1).Format("2006-01-02")
	dayProof := w.Commit()
	mstRoot, _, err := w.CommitMST(dayProof)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle sum tree")
		return
	}
	objects := w.state.IterAllObject()
	mtRoot, err := w.CommitMT(objects)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle tree")
		return
	}
	err = w.db.DiskDB().Put([]byte("lastday"), []byte(w.date))
	if err != nil {
		log.WithError(err).Error("failed to update last day")
		return
	}
	txHash, err := w.rpc.SubmitProofs(w.date, mstRoot, mtRoot)
	if err != nil {
		log.WithError(err).Error("submit proofs")
		return
	}
	log.WithField("tx_hash", txHash.Hex()).Info("submit proofs")

	log.Info("process day done")
}

func (w *Witness) Commit() (dayProofs map[common.Address]*witnessv1.ValidatedProof) {
	st := time.Now()
	proof := w.RefreshPendingProof()
	for miner, proof := range proof {
		err := w.SealProof(miner, proof)
		if err != nil {
			return
		}
	}
	root, err := w.state.Commit()
	if err != nil {
		return
	}
	log.WithFields(log.Fields{"root": root.Hex(), "cost": time.Since(st).String()}).Info("commit state root")
	err = w.db.DiskDB().Put([]byte(fmt.Sprintf("sroot:%s", w.date)), root.Bytes())
	if err != nil {
		log.WithError(err).Error("failed to update state root")
		return
	}
	return proof
}

func (w *Witness) SealProof(miner common.Address, proof *witnessv1.ValidatedProof) (err error) {
	minerObject := w.state.GetMinerObject(miner)
	minerObject.Workload += proof.Workload
	minerObject.Balance += proof.Balance
	err = w.state.UpdateMinerObject(miner, minerObject)
	if err != nil {
		return err
	}
	log.WithFields(log.Fields{
		"miner":        miner.Hex(),
		"old_workload": minerObject.Workload - proof.Workload,
		"new_workload": minerObject.Workload,
		"old_balance":  minerObject.Balance - proof.Balance,
		"new_balance":  minerObject.Balance,
	}).Info("seal proof")
	return err
}

func (w *Witness) RefreshPendingProof() (proof map[common.Address]*witnessv1.ValidatedProof) {
	w.Lock()
	defer w.Unlock()
	proof = make(map[common.Address]*witnessv1.ValidatedProof)

	for miner, pendingProof := range w.pendingProof {
		proof[miner] = pendingProof
	}
	w.pendingProof = make(map[common.Address]*witnessv1.ValidatedProof)
	return
}

// CommitMST commit workload of per day
func (w *Witness) CommitMST(proofMap map[common.Address]*witnessv1.ValidatedProof) (root common.Hash, sum *big.Int, err error) {
	if len(proofMap) == 0 {
		return common.Hash{}, big.NewInt(0), nil
	}
	rawKeys := make([]common.Address, 0)
	keys := make([][32]byte, len(proofMap))
	dbKey := make([]byte, 0)
	dbVal := make([]byte, 0)
	vals := make([]*big.Int, len(proofMap))
	for key := range proofMap {
		rawKeys = append(rawKeys, key)
	}
	sort.Sort(util.AddressSlice(rawKeys))
	for i, key := range rawKeys {
		keys[i] = crypto.Keccak256Hash(key.Bytes())
		dbKey = append(dbKey, keys[i][:]...)
		vals[i] = new(big.Int).SetUint64(proofMap[key].Workload)
		dbVal = append(append(dbVal, vals[i].Bytes()...), []byte(":")...)
	}

	st := time.Now()

	nodes, sums := tree.ComputeMST(keys, vals)

	err = w.db.DiskDB().Put([]byte(fmt.Sprintf("mstroot:%s", w.date)), nodes[0][:])
	if err != nil {
		log.Error(err)
		return
	}
	err = w.db.DiskDB().Put([]byte(fmt.Sprintf("mstsum:%s", w.date)), sums[0].Bytes())
	if err != nil {
		log.Error(err)
		return
	}

	err = w.db.DiskDB().Put([]byte(fmt.Sprintf("mstk:%s", w.date)), dbKey)
	if err != nil {
		return
	}
	err = w.db.DiskDB().Put([]byte(fmt.Sprintf("mstv:%s", w.date)), dbVal)
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"root": common.Hash(nodes[0]).Hex(),
		"sum":  sums[0].String(),
		"cost": time.Since(st).String(),
	}).Info("commit MST root")

	return nodes[0], sums[0], nil
}

// CommitMT commit all workload
func (w *Witness) CommitMT(objects []*witnessv1.MinerObject) (root common.Hash, err error) {
	merkleProofs := make(tree.Proofs, 0)
	dbProofs := make([]byte, 0)
	for _, object := range objects {
		payload := append(common.HexToAddress(object.Miner).Bytes(), common.LeftPadBytes(big.NewInt(0).SetUint64(object.Workload).Bytes(), 32)...)
		_proof := crypto.Keccak256Hash(payload)
		merkleProofs = append(merkleProofs, _proof)
		dbProofs = append(dbProofs, _proof[:]...)
	}

	mtTree, err := tree.NewMerkleTree(merkleProofs)
	if err != nil {
		return
	}
	root = mtTree.GetRoot()

	st := time.Now()
	err = w.db.DiskDB().Put([]byte(fmt.Sprintf("mtroot:%s", w.date)), root.Bytes())
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"k": fmt.Sprintf("mtroot:%s", w.date),
		"v": root.String(),
	}).Debug()

	err = w.db.DiskDB().Put([]byte(fmt.Sprintf("mtk:%s", w.date)), dbProofs)
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"k":        fmt.Sprintf("mtk:%s", w.date),
		"v_length": len(dbProofs),
	}).Debug()
	w.mtTreeCache[w.date] = mtTree

	log.WithFields(log.Fields{
		"root": root.Hex(),
		"cost": time.Since(st).String(),
	}).Info("commit MT root")

	return
}

func (w *Witness) Ticker() {
	duration := time.Duration(24) * time.Hour
	executionTime := time.Date(time.Now().Year(), time.Now().Month(), time.Now().Day(), 0, 0, 0, 0, time.UTC).Add(duration)
	// executionTime := time.Now().Add(time.Second * 30)
	waitTime := executionTime.Sub(time.Now())
	timer := time.NewTimer(waitTime)
	log.WithField("wait_time", waitTime.String()).Info("prepare commit task")
	for {
		<-timer.C
		w.ProcessDay()
		executionTime = executionTime.Add(duration)
		waitTime = executionTime.Sub(time.Now())
		timer.Reset(waitTime)
		log.WithField("wait time", waitTime.String()).Info("prepare commit task")
	}
}

func (w *Witness) Mock() {
	addresses := make([]common.Address, 10)
	rand.Seed(time.Now().Unix())
	for i := 0; i < 10; i++ {
		b20 := make([]byte, 20)
		rand.Read(b20)
		addresses[i] = common.BytesToAddress(b20)
	}

	for {
		time.Sleep(time.Second)
		w.AddPendingProof(addresses[rand.Intn(10)], []*witnessv1.ValidatedProof{
			{
				Workload: uint64((rand.Int63n(9) + 1) * 100),
			},
		})
	}
}

func (w *Witness) UpdateAddress() {
	ticker := time.NewTicker(time.Minute * 10)
	defer ticker.Stop()
	log.Info("start update address task")
	for {
		addrs, err := w.rpc.GetContainerAddresses()
		if err != nil {
			log.WithError(err).Error("failed to get container addresses")
		} else {
			w.containerAddresses = addrs
		}
		addrs, err = w.rpc.GetNMAddresses()
		if err != nil {
			log.WithError(err).Error("failed to get NM addresses")
		} else {
			w.nmAddresses = addrs
		}
		log.WithFields(log.Fields{
			"container_count": len(w.containerAddresses),
			"nm_count":        len(w.nmAddresses),
		}).Info("store contract update address")
		<-ticker.C
	}
}

func (w *Witness) IsContainerAddress(address common.Address) bool {
	for _, addr := range w.containerAddresses {
		if addr.Hex() == address.Hex() {
			return true
		}
	}
	return false
}

func (w *Witness) IsNMAddress(address common.Address) bool {
	for _, addr := range w.nmAddresses {
		if addr.Hex() == address.Hex() {
			return true
		}
	}
	return false
}

func (w *Witness) GetMerkleProof(address common.Address) (workload uint64, proofs []common.Hash) {
	w.Lock()
	cacheTree, ok := w.mtTreeCache[w.date]
	w.Unlock()
	if !ok {
		// todo
	}

	object := w.state.GetMinerObject(address)
	if object == nil {
		return 0, nil
	}

	payload := append(address.Bytes(), big.NewInt(0).SetUint64(object.Workload).Bytes()...)
	leaf := crypto.Keccak256Hash(payload)
	proofs, err := cacheTree.GetProof(leaf)
	if err != nil {
		log.WithError(err).Error("failed to get merkle proof")
		return 0, make([]common.Hash, 0)
	}
	return object.Workload, proofs
}

func (w *Witness) GetPendingWorkload(address common.Address) (workload, globalWorkload uint64) {
	w.Lock()
	proof, ok := w.pendingProof[address]
	w.Unlock()
	if !ok {
		return 0, w.pendingWorkload
	}
	return proof.Workload, w.pendingWorkload
}

func (w *Witness) LoadMerkleProof(date string) {
	merkleTreeKey := fmt.Sprintf("mtk:%s", date)
	data, err := w.db.DiskDB().Get([]byte(merkleTreeKey))
	if err != nil {
		log.WithError(err).Error("failed to load merkle proof")
		return
	}

	proofs := make([]common.Hash, len(data)/32)
	for i := 0; i < len(data)/32; i++ {
		copy(proofs[i][:], data[i*32:(i+1)*32])
	}

	mTree, err := tree.NewMerkleTree(proofs)
	if err != nil {
		log.WithError(err).Error("failed to load merkle proof")
		return
	}
	w.Lock()
	w.mtTreeCache[date] = mTree
	w.Unlock()
	log.WithFields(log.Fields{"date": date, "root": mTree.GetRoot().Hex()}).Info("load merkle tree")
	return
}
